'use strict';

const { setTimeout } = require('timers/promises');
const { expect } = require('chai');
const sinon = require('sinon');
const rewire = require('rewire');
const MemoryWatcher = require('../lib/memory-watcher');

const mockMemoryGlobals = (rewiredModule, totalMemory, usedMemory) => {
  const mocks = [
    rewiredModule.__set__('os', { totalmem: () => totalMemory * 1024 * 1024 }),
    rewiredModule.__set__('process', { memoryUsage: { rss: () => usedMemory * 1024 * 1024 }, on: sinon.stub() })
  ];
  return () => {
    mocks.forEach((revert) => revert());
  };
};

describe('MemoryWatcher', function () {
  describe('constructor', function () {
    beforeEach(function () {

    });

    it('returns a MemoryWatcher instance', function () {
      expect(new MemoryWatcher()).to.be.an.instanceOf(MemoryWatcher);
    });

    it('defaults "liveness" property to "false"', function () {
      const instance = new MemoryWatcher();
      expect(instance.liveness).to.be.false;
    });

    it('defaults "readiness" property to "false"', function () {
      const instance = new MemoryWatcher();
      expect(instance.readiness).to.be.false;
    });
  });

  describe('start', function () {
    let MockableMemoryWatcher;

    beforeEach(function () {
      MockableMemoryWatcher = rewire('../lib/memory-watcher');
    });

    it('logs memory statistics on instantiation and config on startup', async function () {
      const logger = {
        info: sinon.stub(),
        silly: sinon.stub()
      };
      const instance = new MockableMemoryWatcher({ logger, healthyInterval: 1 });
      expect(logger.info.callCount).to.equal(1);
      expect(logger.info.firstCall.args[0]).to.equal('Service startup memory statistics');
      instance.start();
      await setTimeout(10);
      instance.stop();
      await setTimeout(10);
      expect(logger.info.callCount).to.equal(2);
      expect(logger.info.secondCall.args[0]).to.equal('Starting Memory Watcher with the following config');
    });

    it('sets "liveness" to "true" and "readiness" to "true" when current memory usage is below all thresholds', async function () {
      const livenessThresholdOffset = 10;
      const readinessThresholdOffset = 20;
      const totalMemory = 50;
      const usedMemory = totalMemory - readinessThresholdOffset - 1;

      const revert = mockMemoryGlobals(MockableMemoryWatcher, totalMemory, usedMemory);

      const instance = new MockableMemoryWatcher({
        healthyInterval: 1,
        unhealthyInterval: 1,
        livenessThresholdOffset,
        readinessThresholdOffset
      });
      instance.start();
      await setTimeout(10);
      expect(instance.liveness).to.be.true;
      expect(instance.readiness).to.be.true;
      instance.stop();
      await setTimeout(10);
      revert();
    });

    it('sets "liveness" to "true" and "readiness" to "false" when current memory usage exceeds readiness threshold', async function () {
      const livenessThresholdOffset = 10;
      const readinessThresholdOffset = 20;
      const totalMemory = 50;
      const usedMemory = totalMemory - readinessThresholdOffset + 1;

      const revert = mockMemoryGlobals(MockableMemoryWatcher, totalMemory, usedMemory);

      const instance = new MockableMemoryWatcher({
        healthyInterval: 1,
        unhealthyInterval: 1,
        livenessThresholdOffset,
        readinessThresholdOffset
      });
      instance.start();
      await setTimeout(10);
      expect(instance.liveness).to.be.true;
      expect(instance.readiness).to.be.false;
      instance.stop();
      await setTimeout(10);
      revert();
    });

    it('sets "liveness" to "false" and "readiness" to "false" when current memory usage exceeds liveness threshold', async function () {
      const livenessThresholdOffset = 10;
      const readinessThresholdOffset = 20;
      const totalMemory = 50;
      const usedMemory = totalMemory - livenessThresholdOffset + 1;

      const revert = mockMemoryGlobals(MockableMemoryWatcher, totalMemory, usedMemory);

      const instance = new MockableMemoryWatcher({
        healthyInterval: 1,
        unhealthyInterval: 1,
        livenessThresholdOffset,
        readinessThresholdOffset
      });
      instance.start();
      await setTimeout(10);
      expect(instance.liveness).to.be.false;
      expect(instance.readiness).to.be.false;
      instance.stop();
      await setTimeout(10);
      revert();
    });

    it('resets "liveness" and "readiness" back to "true" when current memory usage drops below thresholds after exceeding them', async function () {
      const livenessThresholdOffset = 10;
      const readinessThresholdOffset = 20;
      const totalMemory = 50;
      const usedMemory = totalMemory - livenessThresholdOffset + 1;

      const revert = mockMemoryGlobals(MockableMemoryWatcher, totalMemory, usedMemory);

      const instance = new MockableMemoryWatcher({
        healthyInterval: 1,
        unhealthyInterval: 1,
        livenessThresholdOffset,
        readinessThresholdOffset
      });
      instance.start();
      await setTimeout(10);
      expect(instance.liveness).to.be.false;
      expect(instance.readiness).to.be.false;
      mockMemoryGlobals(MockableMemoryWatcher, totalMemory, totalMemory - readinessThresholdOffset - 1);
      await setTimeout(10);
      expect(instance.liveness).to.be.true;
      expect(instance.readiness).to.be.true;
      instance.stop();
      await setTimeout(10);
      revert();
    });
  });
});
